import numpy as np
from . import QUIC
from tqdm import tqdm

def graph_greedy_search(field_inst, proxy_inst, target_FF, target_FP, target_PP = 0.0, N=30, maxit=500):
    '''
    Solves the graphical lasso problem using the QUIC algorithm.
    Uses a greedy approach to find a graph with a given target sparsity in the
    FF and FP parts of the adjacency matrix. The PP part is assumed diagonal.
               
    Parameters
    ----------
    field_inst: matrix, n x p1 (time x space) 
        Climate field over the instrumental period. Cannot have missing values. 
    proxy_inst: matrix, n x p2 (time x space)
        Proxy field over the instrumental period. Cannot have missing values. 
    target_FF: float in [0,100] 
        Target sparsity of the in-field part of the graph (percentage)
    target_FP: float in [0,100] 
        Target sparsity of the field/proxy part of the graph (percentage)
    target_PP: float in [0,100] 
        Target sparsity of the proxy/proxy part of the graph 
        (this is a dummy argument for back-compatibility ; internally set to 0)    
    N: int
        Number of graphs to consider (Default = 30)
    maxit: int
        maximum number of iterations of the algorithm (Default = 500)
        
    Returns
    -------
    adj_prec: matrix, (p1+p2) x (p1+p2)
        Estimated adjacency matrix (graph) of the total data matrix
    sparsity_prec: float
        Sparsity of the estimated graph
        
    References
    ----------
    Hsieh et al., "QUIC: Quadratic Approximation for Sparse Inverse
		Covariance Estimation", Journal of Machine Learning Research 15 (2014) 2911-2947.     
    
    D. Guillot. B.Rajaratnam. J.Emile-Geay. "Statistical paleoclimate reconstructions via Markov random fields." 
       Ann. Appl. Stat. 9 (1) 324 - 352, March 2015.    
    '''
    # TODO: ask Dominique to rewrite this for FF, FP search only.
    
    if np.isnan(field_inst).any() or np.isnan(proxy_inst).any():
        raise ValueError("Error: this method requires the data matrices to be gap-free")
        
    print("Solving graphical LASSO using greedy search")

    X = np.hstack((field_inst, proxy_inst))
    ind_F = range(field_inst.shape[1])
    ind_P = range(field_inst.shape[1],field_inst.shape[1]+proxy_inst.shape[1])
    S = np.corrcoef(X.T)
    p = S.shape[0]
    pen = np.zeros((p,p))   # Penality matrix
    adj = np.zeros((p,p))
    adj_prec = np.zeros((p,p))

    rho_max = (S - np.diag(np.diag(S))).max()
    rho_min = 0.1*rho_max

    O = np.eye(p)
    W = np.eye(p)

    rhos = np.linspace(rho_min, rho_max, N)

    c_FF = N-1
    c_FP = N-1
    c_PP = N-1

    it = 0
    stop = 0

    visited = np.array([])   # Visited positions
    sparsity = [0.0,0.0,0.0]
    

    # print("Iter    FF      FP      PP\n")
    sparsity_FF, sparsity_FP, sparsity_PP = 0, 0, 0
    
    # print("%1.3d  %6.3f  %6.3f  %6.3f" % (it, sparsity_FF, sparsity_FP, sparsity_PP))
    pbar = tqdm(range(maxit), desc=f'graph_greedy_search | FF: {sparsity_FF:6.3f}; FP: {sparsity_FP:6.3f}; PP: {sparsity_PP:6.3f}')
    for it in pbar:
        pbar.set_description(f'graph_greedy_search | FF: {sparsity_FF:6.3f}; FP: {sparsity_FP:6.3f}; PP: {sparsity_PP:6.3f}')
    # while (stop == 0) & (it < maxit):
        sparsity_prec = np.copy(sparsity)
        adj_prec = np.copy(adj)
        # Build penalty matrix
        pen[np.ix_(ind_F,ind_F)] = rhos[c_FF]
        pen[np.ix_(ind_F,ind_P)] = rhos[c_FP]
        pen[np.ix_(ind_P,ind_F)] = rhos[c_FP]
        pen[np.ix_(ind_P,ind_P)] = rhos[c_PP]

        [W, O] = QUIC.QUIC(S, pen)
        # Compute sparsity of the different parts (TODO)
        sparsity_FF, sparsity_FP, sparsity_PP, adj = sp_level_3(O, ind_F, ind_P)
        sparsity = [sparsity_FF, sparsity_FP, sparsity_PP]

        if (sparsity_FF < target_FF):
            c_FF = max(0,c_FF-1)

        if (sparsity_FP < target_FP):
            c_FP = max(0,c_FP-1);

        if (sparsity_PP < target_PP):
            c_PP = max(0,c_PP-1)

        it = it + 1
        # Print current iteration information
        # print("%1.3d  %6.3f  %6.3f  %6.3f" % (it, sparsity_FF, sparsity_FP, sparsity_PP))
        c_point = c_FF + c_FP*N + c_PP*N**2

        if (c_point in visited):
            stop = 1
        else:
            visited = np.append(visited, c_point)
    
    # for consistency with neighborhood graphs, set PP to 1s
    adj_prec[np.ix_(ind_P, ind_P)] = np.eye(len(ind_P))
    # print(type(adj_prec), type(sparsity_prec))

    return adj_prec, sparsity_prec


from numpy import cos, sin, arcsin, sqrt, radians
import itertools


def neighbor_graph(lonlat, ind_F, ind_P, cutoff_radius = 1000):
    '''
    Constructs a neighborhood graph. Vertices are connected 
    if and only if they are at less than a given cutoff radius.
    
    Parameters:
    ----------
    lonlat: matrix, p x 2
        Matrix where each row contains the (longitude, latititude) of a location.
    ind_F: vector
        Vector containing field gridpoint indices
    ind_P: vector 
        Vector containing the indices of the proxies
    cutoff_radius: float, positive
        Radius of the neighborhood graph
        
    Returns: 
    --------
    adj: matrix
        Adjacency matrix of the neighborhood graph
    sparsity: float
        Sparsity level of the adjacency matrix
    '''
    print("Estimating graph using neighborhood method")

    D = dMat(lonlat)
    adj = (D <= cutoff_radius).astype(int)           # replace 0s or 1s
    adj[np.ix_(ind_P, ind_P)] = np.eye(len(ind_P)) # set pp to 1s
    sparsity = graph_sparsity(adj, ind_F, ind_P)

    return [adj, sparsity]


def great_circle_distance(lon1, lat1, lon2, lat2):
    '''
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)

    Parameters:
    -----------
    lat1: float
        Latitude of first location (degrees)
    lon1: float
        Longitude of first location (degrees)
    lat2: float
        Latitude of second location (degrees)
    lon2: float
        Longitude of second location (degrees)
        
    Returns:
    --------
    km: float
        Great circle distance between the two locations (in kilometers)
    '''
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * arcsin(sqrt(a))

    # 6367 km is the radius of the Earth
    km = 6367 * c
    return km

def dMat(lonlat):
    '''
    Computes a matrix containing the distance between every pair of points 
    
    Parameters:
    -----------
    lonlat: matrix, p x 2
        Matrix where each row contains the (longitude, latititude) of a location.
        
    Returns:
    --------
    D: matrix, p x p
        Matrix whose (i,j)-th entry contains the great circle distance (in km) between location i 
        and location j.
    '''
    N = lonlat.shape[0]
    D = np.zeros((N,N))
    for pair in itertools.product(range(N),repeat=2):
        D[pair] = great_circle_distance(lonlat[pair[0],0],lonlat[pair[0],1],lonlat[pair[1],0],lonlat[pair[1],1])

    return D

def graph_sparsity(adj, ind_F, ind_P):
    '''
    Computes the sparsity level of the graph specified by an adjacency matrix in the 
    in-field, field/proxy, and proxy/proxy parts of the matrix.

    Parameters:
    -----------
    adj: matrix
        Adjacency matrix of the graph
    ind_F: vector
        Vector containing the indices of the field locations
    ind_P: vector
        Vector containing the indices of the proxies
        
    Returns:
    --------
    sparsity_FF: float
        Sparsity level of the in-field part of the graph
    sparsity_FP: float 
        Sparsity level of the field/proxy part of the graph
    sparsity_PP: float
        Sparsity level of the proxy/proxy part of the graph
    '''
    n_T = len(ind_F)
    n_P = len(ind_P)
    sparsity_FF = (adj[np.ix_(ind_F, ind_F)] - np.eye(n_T)).sum()/(n_T**2-n_T)  # n**0.5 - n 
    sparsity_FP = adj[np.ix_(ind_F, ind_P)].sum()/(n_T*n_P)
    sparsity_PP = 0.0

    return [sparsity_FF, sparsity_FP, sparsity_PP]

def adj_matrix(S, tol=1e-3):
    '''
    Computes the adjacency matrix of the graph associated to a matrix
    via thresholding

    Parameters:
    -----------
    S: matrix
        Matrix to compute the graph of
    tol: thresholding value (Default = 1e-3)

    Returns:
    --------
    adj: matrix
        Adjacency matrix of the graph associated to S.
    '''
    adj = (abs(S) > tol).astype(int)
    adj = adj - np.diag(np.diag(adj))
    return adj

def sp_level_3(O, ind_F, ind_P):
    '''
    Computes the sparsity level (percentage of nonzero entries) 
    of a matrix in the in-field, 
    field/proxy, and proxy/proxy part of the matrix O.

    Parameters
    ----------
    O: matrix
        Matrix to compute the sparsity level of
    ind_F: vector of int
        Indices of the field locations
    ind_P: vector of int
        Indices of the proxy locations
        
    Returns
    -------
    level_FF: sparsity level of the in-field block of the matrix
    level_FP: sparsity level of the field/proxy block of the matrix
    level_PP: sparsity level of the proxy/proxy block of the matrix
    adj: adjacency matrix associated to the matrix O
    '''
    p = O.shape[0]
    d = np.diag(O)
    dinv = np.sqrt(d)**(-1)
    D = np.diag(dinv)
    O_scaled = D.dot(O).dot(D)
    adj = adj_matrix(O_scaled)
    p_FF = len(ind_F)
    p_PP = len(ind_P)

    adj_FF = adj[np.ix_(ind_F, ind_F)]
    level_FF = adj_FF.sum().astype(float) / (p_FF*(p_FF-1))*100

    adj_FP = adj[np.ix_(ind_F, ind_P)]
    level_FP = adj_FP.sum().astype(float) / (p_FF*p_PP)*100

    adj_PP = adj[np.ix_(ind_P, ind_P)]
    level_PP = adj_PP.sum().astype(float) / (p_PP*(p_PP-1))*100

    adj = adj + np.eye(p)
    return [level_FF, level_FP, level_PP, adj]



import numpy as np
from scipy.sparse.linalg import eigs
from scipy.interpolate import interp1d
from scipy.optimize import minimize_scalar
#####

'''
Collections of functions to perform Ridge regression, with 
generalized cross-validation to choose the Ridge regularization 
parameter.
'''

def gcvfctn (h, d, fc2, trS0, dof0):
	'''
		h:  regularization parameter,
		d:  column vector of eigenvalues of cov(X),
	  fc2:  row sum of squared Fourier coefficients, fc2=sum(F.^2, 2),
	 trS0:  trace(S0) = Frobenius norm of generic part of residual matrix,
	 dof0:  degrees of freedom in estimate of residual covariance
			matrix when regularization parameter is set to zero
	'''

	filfac = (h**2) / (d + h**2)
	filfac = filfac.reshape(len(filfac),1)
	A = dof0 + sum(filfac)
	g = (sum((filfac**2) * fc2) + trS0 ) / A**2
	return g

#####

def gcvridge(F, d, trS0, n, r, trSmin, minvarfrac = 0):
	'''
		F:  the matrix of Fourier coefficients,
		d:  column vector of eigenvalues of X'*X/n,
	 trS0:  trace(S0) = trace of generic part of residual 2nd moment matrix,
		n:  number of degrees of freedom for estimation of 2nd moments,
		r:  number of nonzero eigenvalues of X'*X/n,
	trSmin:  minimum of trace(S_h) to construct approximate lower bound
			on regularization parameter h (to prevent GCV from choosing
			too small a regularization parameter).
	'''
	# Ensure d is in double
	d = np.atleast_1d(d)

	# Check sizes of input matrices
	if len(d) < r:
		print("All nonzero eigenvalues must be given.")
		return

	if minvarfrac < 0  or minvarfrac > 1:
		print("error: minvarfrac must be in [0,1].")
		return

	p = F.shape[0]

	if p < r:
		print("F must have at least as many rows as there are nonzero eigenvalues d")
		return

	# Row sum of squared Fourier coefficients

	#fc2 = np.sum(F**2, axis = 1)
	fc2 = F**2
	fc2 = np.reshape(fc2,(len(fc2),1))         # Reshape

	# Accuracy of regularization parameter
	h_tol  = 0.2/(np.sqrt(n))

	# Heuristic upper bound on regularization parameter
	varfrac = d.cumsum(axis=0)/d.sum(axis=0)

	if minvarfrac > varfrac.min():
		func = interp1d(varfrac, d)
		d_max = func(minvarfrac)
		h_max = np.sqrt(d_max)
	else:
		h_max = np.sqrt(d.max()) / h_tol

	# Heuristic lower bound on regularization parameter

	if trS0 > trSmin:
		h_min = np.finfo(float).eps**0.5
	else:
		rtsvd = np.zeros((r, 1))
		rtsvd[r-1] = trS0

		for j in range(r-2,0,-1):
			rtsvd[j] = rtsvd[j+1] + fc2[j+1]

		rmin = (abs(rtsvd - trSmin)).argmin()
		h_min = (max(d[rmin], d.min()/n))** 0.5

	# find minimizer of GCV function
	if h_min < h_max:
		fun = lambda x: gcvfctn(x,d[0:r], fc2[0:r], trS0, n-r)
		soln = minimize_scalar(fun, bounds = (h_min, h_max), method='bounded')
		h_opt  = soln['x']
	else:
		print("Upper bound on regularization parameter smaller than lower bound.")
		h_opt  = h_min

	return h_opt

#####

def peigs (A, rmax):
	'''
	'''

	# check number of input arguments
	[m,n] = A.shape

	if rmax > min(m,n):
		rmax  = min(m,n)  # rank cannot exceed size of A

	if rmax < min(m, n)/10.0:
		d,V = eigs(A, rmax)
	else:
		d,V = np.linalg.eig(A)

	# ensure eigenvalues are monotonically decreasing
	I = np.argsort(d)
	I = I [::-1]
	d = d[I]
	V = V[:,I]

	# estimate number of positive eigenvalues of A
	d_min = d.max() * max(m,n) * np.finfo(float).eps

	r = (d > d_min).sum()

	# discard eigenpairs with eigenvalues that are close to or less than zero
	d = d[0:r]
	V = V[:,0:r]                # V = V(:, 1:r)

	return [V,d,r]

#####

def iridge(C, xind, yind, dof, relvar_res = 5e-2):
	'''
	'''
	Cxx = C[np.ix_(xind, xind)]
	Cyy = C[np.ix_(yind, yind)]
	Cxy = C[np.ix_(xind, yind)]

	px = Cxx.shape[0]
	py = Cyy.shape[0]

	# eigendecomposition of Cxx
	rmax = min(dof, px)     # maximum possible rank of Cxx

	[V, d, r] = peigs(Cxx, rmax)
	V = V.real # In case Cxx is not perfectly symmetric numerically
	d = d.real

	# Fourier coefficients
	Vt = V.conj().T
	F = (np.tile((1/(d**0.5)).reshape(len(d),1),px) * Vt).dot(Cxy)

	if dof > r:
		S0  = Cyy - F.conj().T.dot(F)
	else:
		S0  = np.zeros((py, py))

	# approximate minimum squared residual
	trSmin = (relvar_res)* (np.diag(Cyy))

	# initialize output
	h   = np.zeros((py, 1))
	B   = np.zeros((px, py))

	S  = np.zeros((py, py))
	peff = np.zeros((py, 1))

	for k in range (py):
		
		h[k] = gcvridge(F[:,k], d, S0[k,k], dof, r, trSmin[k], relvar_res)
		# k-th column of matrix of regression coefficients
		B[:,k]  = V.dot( (d**0.5 / (d + h[k]**2) * F[:,k]))

		# if S_out     assemble estimate of covariance matrix of residuals
		for j in range(k+1): # k starts at 0
			diagS  = ( h[j]**2 * h[k]**2 ) / ( (d + h[j]**2) * (d + h[k]**2))            
			S[j,k] = S0[j,k] + (F[:,j].conj().T).dot(diagS * F[:,k])
			S[k,j] = S[j,k]
		# if peff_out  effective number of adjusted parameters in this column
		# of B: peff = trace(Mxx_h Cxx)
		peff[k]  = sum(d / (d + h[k]**2))

	return [B,S,h,peff]


####

# Testing
def test():
	C = np.array([[1,2,3],[4,5,6],[7,8,9]])
	print(iridge(C, [0,1], [2], 2))
#[array([[ 0.24669516],[ 0.6739837 ]]), array([[ 2.76176439]]), array([[ 1.68022783]]), array([[ 0.69601762]])]
'''
>> [B,S,h,peff]=iridge(Caa, Cmm, Cam,2,[])

B =

    0.2467
    0.6740


S =

    2.7618


h =

    1.6802


peff =

    0.6960

'''

#    A = np.array([[1,2,3,4,5],[6,7,8,9,0],[1,3,5,7,9],[9,7,4,2,3],[1,2,3,4,5]])
#    print(iridge(A,[0,1,2],[3,4],2))

#[array([[ 0.00403285,  0.00353936],[ 0.02064954, -0.00077429],[ 0.00474236,  0.00794145]]), 
#      array([[ 25.16699546, -22.16105325],[-22.16105325,  50.74469669]]), 
#      array([[ 24.77492727], [ 24.78841235]]), 
#      array([[ 0.02078152], [ 0.02075934]])]

'''
>> [B,S,h,peff]=iridge(Caa, Cmm, Cam,2,[])

B =

    0.0040    0.0035
    0.0206   -0.0008
    0.0047    0.0079


S =

   25.1676  -22.1609
  -22.1609   50.7449


h =

   24.7914  # 24.7749
   24.8003  # 24.7884


peff =

    0.0208
    0.0207

'''



from .GraphEstimation import dMat
from .GraphEstimation import graph_sparsity
from .GraphEstimation import graph_greedy_search

import numpy as np

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
from matplotlib.colors import BoundaryNorm


class Graph:
    ''' The object for a graph '''

    def __init__(self, lonlat, field, proxy):
        '''
        Args:
            lonlat (numpy.array): the location array with shape (num_grid + num_proxy, 2), where the 2nd dimension is aranged as [longitude, latitude]
            field (numpy.array): the climate field with shape (num_time, num_grid)
            proxy (numpy.array): the proxy matrix with shape (num_time, num_proxy)
        '''
        self.lonlat = lonlat
        self.field = field
        self.proxy = proxy

    def calc_distance(self):
        ''' Calculate the distance matrix

        Attributes:
            dist_mat (numpy.array): the great circle distance matrix
        '''
        
        self.dist_mat = dMat(self.lonlat)
        

    def neigh_adj(self, cutoff_radius = None, n_neighbors = None):
        ''' Calculate the adjacency matrix for a neighborhood graph

        The neighborhood is controlled either/and by the cutoff radius or the 
        number of neigbors.  If both are specified, the beighborhood is 
        comprised of the `n_neighbors` closest points within a distance of 
        `cutoff_radius`. An error is raised if neither method is specified.

        Parameters
        ----------
        cutoff_radius : float
            the great circle distance in unit of [km] for neighbor search.
            Default: None
        n_neighbors : int
            number of closest neighbors (Default: None)

        Returns
        -------
        adj (numpy.array): the adjacency matrix (num_grid + num_proxy, num_grid + num_proxy)
        '''
        
        if not hasattr(self, 'dist_mat'):
            self.calc_distance()
            
        # make distance matrix upper triangular
        D = self.dist_mat.copy()
        D[np.tril_indices_from(D)] = np.nan
        p = D.shape[0] # matrix size
        
        adj   = np.zeros((p,p))   
        ind_F = range(self.field.shape[1])
        ind_P = range(self.field.shape[1], p)
        
        if cutoff_radius is not None and n_neighbors is not None:        
            D[D>cutoff_radius] = np.Inf
            for j in range(p):       
                idx = np.argsort(D[j,:])
                nonnan = np.sum(~np.isnan(D[j,idx]))
                adj[j,idx[:np.min([nonnan,n_neighbors])]] = 1

        elif cutoff_radius is not None and n_neighbors is None:   
            adj = (self.dist_mat <= cutoff_radius).astype(int)  # Booleans as 0s or 1s
            
        elif cutoff_radius is None and n_neighbors is not None:
            for j in range(p):       
                idx = np.argsort(D[j,:])
                nonnan = np.sum(~np.isnan(D[j,idx]))
                adj[j,idx[:np.min([nonnan,n_neighbors])]] = 1
        else:
            print('No rule was specified to create the neighborhood')
        
        # symmetrize adjacency matrix
        adj = np.where(adj,adj,adj.T)  # h/t https://stackoverflow.com/a/58718913/4984978
        
        # set PP block to identity
        adj[np.ix_(ind_P, ind_P)] = np.eye(len(ind_P)) 
        # restore diagnonal over ind_F
        for j in ind_F:   
            adj[j,j] = 1
        
        self.adj = adj
        self.cutoff_radius = cutoff_radius
        self.n_neighbors = n_neighbors
        self.sparsity = graph_sparsity(adj, ind_F, ind_P)
        
    def glasso_adj(self, target_FF, target_FP):
        ''' Calculates the adjacency matrix using the Graphical LASSO. 
            Uses a QUIC solver and a greedy approach to find a graph 
            with a given target sparsity in the FF and FP parts of 
            the adjacency matrix. The PP part is assumed diagonal.

        Parameters
        ----------
        
        target_FF: float in [0,100] 
            Target sparsity of the in-field part of the graph (percentage)
        target_FP: 
            Target sparsity of the field/proxy part of the graph (percentage)
            
        Returns
        -------
        adj : matrix, (p1+p2) x (p1+p2)
            Estimated adjacency matrix (graph) of the total data matrix
        sparsity:  float x 3
            Sparsity of the estimated graph
            
        References
        ----------
        Hsieh et al., "QUIC: Quadratic Approximation for Sparse Inverse
    		Covariance Estimation", Journal of Machine Learning Research 15 (2014) 2911-2947.     
        
        Guillot, Rajaratnam & Emile-Geay. "Statistical paleoclimate reconstructions 
            via Markov random fields" Ann. Appl. Stat. 9 (1) 324 - 352, March 2015.    

        '''
        adj, sp = graph_greedy_search(self.field, self.proxy, target_FF, target_FP)      
        self.adj = adj
        self.sparsity = sp
        
        
    def plot_adj(self,figsize=(6, 6), clr='C0', ax=None, title=None):
        ''' Plot the adjacency matrix for a neighborhood graph

        Args:
            figsize (tuple) : figure size (optional)
            clr : color of the rectangular patches and text delineating regions of the matrix   


        Attributes:
            adj (numpy.array): the adjacency matrix in shape of (num_grid + num_proxy, num_grid + num_proxy)

        '''
        num_tot = self.adj.shape[1]
        num_grid = self.field.shape[1]
        num_proxy = self.proxy.shape[1]
        
        assert num_tot == num_grid+num_proxy, "matrix dimensions do not add up"
        
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize) # in inches

        ax.imshow(self.adj,cmap="Greys",interpolation="none")
        ax.set_xlabel('Index')
        
        # TODO : better label positions, using rcParams
        
        # plot climate-climate part of the graph
        ax.add_patch(plt.Rectangle((0, 0), num_grid, num_grid, alpha = 0.3,
                                   fc=clr, # face color
                                   ec='none'))  # edge color
        ax.annotate('in-field', color=clr,
                    xy=(num_grid/2, 0),xycoords='data',
                    xytext=(num_grid/2/num_tot, 1.02), textcoords='axes fraction')
        
        # plot climate-proxy part of the graph
        fac= 0.75
        ax.add_patch(plt.Rectangle((num_grid, 0), num_proxy-2, num_grid,
                                   fc='none', ec=clr, linewidth=2))
        ax.annotate('field-proxy', color=clr,
                    xy=(num_grid+fac*num_proxy/2, 0),xycoords='data',
                    xytext=((num_grid+fac*num_proxy/2)/num_tot, 1.02), textcoords='axes fraction')
        
        # plot proxy-proxy part of the graph
        ax.add_patch(plt.Rectangle((num_grid+1, num_grid+1), num_proxy-2, num_proxy,
                                   linewidth=2,ls='--',
                                   fc='none', ec=clr))
                    
        ax.text(1.05*num_tot,num_grid+1.4*num_proxy/2, s='proxy-proxy', 
                rotation='vertical', color = clr)

        if title is not None:
            ax.set_title(title, y=1.1)
        
        if 'fig' in locals():
            return fig, ax
        else:
            return ax
        

    def get_neighbor_locs(self, idx):
        n_gridpts = self.field.shape[1]
        i = idx + n_gridpts

        target_lon, target_lat = self.lonlat[i]
        neighbor_lons = []
        neighbor_lats = []
        neighbor_idx = []
        for j, m in enumerate(self.adj[i, :]):
            if m == 1:
                if list(self.lonlat[j]) != [target_lon, target_lat]:
                    neighbor_lon, neighbor_lat = self.lonlat[j]
                    neighbor_lons.append(neighbor_lon)
                    neighbor_lats.append(neighbor_lat)
                    neighbor_idx.append(j)

        return target_lon, target_lat, neighbor_lons, neighbor_lats, neighbor_idx

    def plot_neighbors(self, idx, figsize=(4, 4), ms=50, title=None, neighbor_clr='r',
                       target_clr='k', edge_clr='w', marker='o', cmap=None, norm=None, ax=None):
        ''' Plot the location of the neighbors according to the adjacency matrix

        Parameters
        ----------

        idx : int
            the index of the target proxy

        figsize : tuple
            the figure size

        title : str
            the title string

        ms : float
            the marker size

        neighbor_clr : str or color list
            the colors for the neighbors

        target_clr : str or color list
            the color for the target proxy

        edge_clr : str or color list
            the color for the edge of the markers

        marker : str
            the marker symbol

        Returns
        ----------

        adj : array
            the adjacency matrix in shape of (num_grid + num_proxy, num_grid + num_proxy)

        '''
        target_lon, target_lat, neighbor_lons, neighbor_lats, _ = self.get_neighbor_locs(idx)

        if ax is None:
            fig = plt.figure(figsize=figsize)
            ax = plt.subplot(projection=ccrs.Orthographic(central_longitude=target_lon, central_latitude=target_lat))

        ax.set_global()
        ax.stock_img()

        if title is None:
            if hasattr(self, 'cutoff_radius'):
                ax.set_title(f'Neighbors (r={self.cutoff_radius} km)')
            else:
                ax.set_title(f'Neighbors of the target proxy')
        else:
            ax.set_title(title)


        transform=ccrs.PlateCarree()
        if cmap is not None:
            ax.scatter(neighbor_lons, neighbor_lats, marker=marker, s=ms, c=neighbor_clr, edgecolor=edge_clr,transform=transform, cmap=cmap, norm=norm)
        else:
            ax.scatter(neighbor_lons, neighbor_lats, marker=marker, s=ms, c=neighbor_clr, edgecolor=edge_clr,transform=transform)
        ax.scatter(target_lon, target_lat, marker=marker, s=ms, c=target_clr, edgecolor=edge_clr, transform=transform)

        if 'fig' in locals():
            return fig, ax
        else:
            return ax


    def plot_neighbors_corr(self, idx, time_idx_range=None, figsize=(5, 5), ms=50, title=None,
        cmap='RdBu_r', target_clr='k', edge_clr='w', marker='o',  levels=np.linspace(-1, 1, 21),
        cbar_pad=0.1, cbar_orientation='horizontal', cbar_aspect=10, cbar_labels=[-1, -0.5, 0, 0.5, 1],
        cbar_fraction=0.15, cbar_shrink=0.5, cbar_title='Correlation', plot_cbar=True, ax=None):
        ''' Plot the location of the neighbors according to the adjacency matrix

        Parameters
        ----------

        idx : int
            the index of the target proxy

        time_idx_range : list
            the list of time indices for correlation calculation

        figsize : tuple
            the figure size

        title : str
            the title string

        ms : float
            the marker size

        neighbor_clr : str or color list
            the colors for the neighbors

        target_clr : str or color list
            the color for the target proxy

        edge_clr : str or color list
            the color for the edge of the markers

        marker : str
            the marker symbol

        plot_cbar : bool
            if True, plot the colorbar

        Returns
        ----------

        adj : array
            the adjacency matrix in shape of (num_grid + num_proxy, num_grid + num_proxy)

        '''
        _, _, _, _, neighbor_idx = self.get_neighbor_locs(idx)
        target_value = self.proxy[:, idx]
        corrs = []
        # print('np.shape(target_value):', np.shape(target_value))
        for i in neighbor_idx:
            neighbor_value = self.field[:, i]
            if time_idx_range is None:
                time_idx_range = ~np.isnan(neighbor_value)
            # print('np.shape(neighbor_value):', np.shape(neighbor_value))
            corr = np.corrcoef(neighbor_value[time_idx_range], target_value[time_idx_range])[1, 0]
            corrs.append(corr)
        # print(corrs)

        cmap = plt.get_cmap(cmap)
        norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)

        if ax is None:
            fig, ax = self.plot_neighbors(
                idx, figsize=figsize, ms=ms, title=title, neighbor_clr=corrs, target_clr=target_clr,
                edge_clr=edge_clr, marker=marker, cmap=cmap, norm=norm)
        else:
            ax = self.plot_neighbors(
                idx, figsize=figsize, ms=ms, title=title, neighbor_clr=corrs, target_clr=target_clr,
                edge_clr=edge_clr, marker=marker, cmap=cmap, norm=norm, ax=ax)

        smap = plt.cm.ScalarMappable(cmap=cmap, norm=norm)

        if plot_cbar:
            cbar = plt.colorbar(smap, ax=ax, orientation=cbar_orientation, pad=cbar_pad, aspect=cbar_aspect, extend='neither', fraction=cbar_fraction, shrink=cbar_shrink)
            cbar.ax.set_title(cbar_title)
            cbar.set_ticks(cbar_labels)

        if 'fig' in locals():
            return fig, ax
        else:
            return ax





import numpy as np
from scipy import linalg
import math
#from . import QUIC
from . import iridge
from .GraphEstimation import graph_greedy_search
from .GraphEstimation import neighbor_graph
import matplotlib.pyplot as plt
from matplotlib import gridspec
from tqdm import tqdm


class GraphEM(object):
    ''' The class for the GraphEM solver

    Implementation of the GraphEM algorithm

    Reference: D. Guillot. B.Rajaratnam. J.Emile-Geay. "Statistical paleoclimate reconstructions via Markov random fields."
               Ann. Appl. Stat. 9 (1) 324 - 352, March 2015.

    Args:
        tol (float): convergence tolerence of EM algorithm (default: 5e-3)
        field_r (numpy.ndarray): reconstructed climate field
        proxy_r (numpy.ndarray): reconstructed proxy matrix
        calib (numpy.ndarray): index of calibration period
        Sigma (numpy.ndarray): covariance matrix of the multivariate normal model
        mu (numpy.ndarray): mean of the multivariate normal model

    '''
    def __init__(self, tol=5e-3):
        ''' Initializes the GraphEM object'''

        self.tol = tol
        self.field_r = []
        self.proxy_r = []
        self.calib = []
        self.Sigma = []
        self.mu = []

    def fit(self, field, proxy, calib, graph=[], lonlat=[], sp_FF=3.0, sp_FP=3.0, sp_PP=3.0, N_graph=30, C0=[], M0=[], maxit=200,
        bootstrap=False, N_boot=20, cutoff_radius=1000, graph_method='neighborhood', estimate_graph=True, save_graphs=False, verbose=False):
        '''
        Estimates the parameters of the GraphEM model and reconstruct the missing values of the climate
        and proxy fields.

        Parameters
        ----------
        field: matrix (time x space)
            Climate field. Missing values stored as "np.nan".
        proxy: matrix (time x space)
            Proxy data. Missing values stored as "np.nan". The time dimension should be
                        the same as the climate field.
        calib: vector
            Vector of indices representing the calibration period.
        graph: NumPy array
            Adjacency matrix of the climate+proxy data matrix. (Default = [], estimated via graph_greedy_search)
        lonlat: matrix ((number of grid points + number of proxies) x 2). Default = []
            Matrix containing the (longitude, latitude) of the climate field grid points and proxy locations. Only
            used if graph_method = 'neighborhood'.
        sp_FF: float
            Target sparsity of the in-field part of the inverse covariance matrix. Only used
               when the graph is estimated by glasso. Default (3.0%)
        sp_FP: float
            Target sparsity of the climate field/proxy part of the inverse covariance matrix. Only used
               when the graph is estimated by glasso. Default (3.0%)
        sp_PP: float
            Target sparsity of the proxy/proxy part of the inverse covariance matrix. Only used
               when the graph is estimated by glasso. Default (3.0%)
        N_graph: int
            Number of graphs to consider in the graph_greedy_search method (Default = 30). Only used
                 if the graph is estimated using glasso.
        C0: matrix
            Initial estimate of the covariance matrix of the total data matrix (climate + proxies). (Default = []).
        M0: vector
            Initial estimate of the mean of the total data matrix (climate + proxies). (Default = []).
        bootstrap: boolean
            Indicates whether or not to produce multiple estimates of the reconstructed values via bootstrapping. (Default = False)
        N_boot: int
            Number of bootstrap samples to use if using the bootstrap method.
        cutoff_radius: float
            Radius of the neighborhood graph. Only used if graph_method = 'neighborhood'. (Default = 1000 km).
        graph_method: "neighborhood" or "glasso"
            Method to use to estimate the graph. Used only if graph = [] or if estimate_graph = True.
        save_graphs: boolean
            Indicates whether or not to save all the graphs return by graph_greedy_search (Default = False).

        '''

        if ~bootstrap:
            self.field = field
            self.proxy = proxy
            self.calib = calib
            self.lonlat = lonlat
            self.graph = graph
            self.N_boot = N_boot

            self.graph_method = graph_method
            self.cutoff_radius = cutoff_radius
            self.sp_FF = sp_FF
            self.sp_FP = sp_FP
            self.sp_PP = sp_PP
            self.N_graph = N_graph

        ind_F = range(field.shape[1])
        ind_P = range(field.shape[1], field.shape[1]+proxy.shape[1])
        X = np.hstack((field, proxy))

        if bootstrap:
            self.bootstrap(N_boot = N_boot, save_graphs = save_graphs)
            self.field_r = self.field_r_all.mean(axis=0)
        else:
            if estimate_graph:
                if graph_method == "neighborhood":
                    print("Computing a neighborhood graph with R = {:4.1f} km".format(self.cutoff_radius))
                    if len(lonlat) == 0:
                        print("Error: you need to specify the longitude/latitude of grid points and proxy locations (lonlat)")
                        return
                    else:
                        [self.graph, self.sparsity] = neighbor_graph(lonlat, ind_F, ind_P, self.cutoff_radius)

                elif graph_method == "glasso":
                    print("Applying the graphical lasso")
                    # Test if there are any missing values in the calibration period
                    if np.any(np.isnan(X[calib,:])):  # Use iridge to reconstruct missing values on the calibration period
                        print("Infilling missing values over the calibration period using RegEM iridge")
                        Xridge, __, __ = self.EM(X[calib,:], [], C0, M0, maxit, use_iridge = True)
                        [self.graph, self.sparsity] = graph_greedy_search(Xridge[:,ind_F], Xridge[:,ind_P], sp_FF, sp_FP, sp_PP, N_graph)
                    else:
                        [self.graph, self.sparsity] = graph_greedy_search(field[calib,:], proxy[calib,:], sp_FF, sp_FP, sp_PP, N_graph)
                else:
                    return "Error: graph can't be generated! Please choose a graph option."

            else:
                if graph == []:
                    return "Error: you need to specify a graph if estimate_graph = False."
                else:
                    if verbose: print("Using specified graph")

            [X,C,M] = self.EM(X, self.graph, C0, M0, maxit, verbose=verbose)
            self.field_r = X[:,ind_F]
            self.proxy_r = X[:,ind_P]
            self.Sigma = C
            self.mu = M

    def EM(self, X, graph, C0 = [], M0 = [], maxit = 200, use_iridge = False, verbose=False):
        '''
        Expectation-Maximization (EM) algorithm with Gaussian Graphical Model

        Parameters
        ----------
        X: matrix (n x p), time x space
            Temperature + Proxy matrix with missing values (np.na)
        graph: matrix (p x p)
            Adjacency matrix of the graph
        C0: matrix (p x p) 
            Initial covariance matrix of the field (Default = [], uses sample covariance matrix)
        M0: vector (p) 
            Initial mean of the field (Default = [], uses sample mean)
        maxit: int
            Maximum number of iteration of the algorithm (Default = 200)
        use_iridge: boolean
            If True, uses Ridge regularization to perform regression (Default = False)
            
        Returns
        -------
        X: matrix (n x p)
            Field with reconstructed values
        C: matrix (p x p) 
            Estimated covariance matrix of the field
        M: vector (p)
            Estimate mean vector of the field
        '''
        
       # Start EM algorithm
        if verbose: print("Running GraphEM:\n")
        [n,p] = X.shape
        
        # Find unique lines
        indmis = np.isnan(X)
        nmis = len(find(indmis))
        (pattern, pattern_ix) = unique_rows(indmis)
        s = pattern.shape
        if len(s) == 1:
            nb_pattern =1
            pattern = pattern.reshape((1,s))
        else:
            nb_pattern = s[0]
            
        if M0 == []:
            # Center data
            Xmask = np.ma.masked_array(X,indmis)
            M = Xmask.mean(axis = 0)
            M = M.filled(np.nan)
        else: 
            M = M0
            
        X = X-M
        X[indmis] = 0.0

        if C0 == []:
            if use_iridge:
                C = X.T.dot(X)/(n-1)
            else:
                C = X.T.dot(X)/(n-1) + 0.1*np.eye(p) # boost rank of C for initial estimate
        else:
            C = C0

        it = 0
        rdXmis = np.inf

        Xmis = np.zeros((n,p))
        dXmis = np.nan

        # if verbose: print("Iter     dXmis     rdXmis\n")
        
        pbar = tqdm(range(maxit), desc=f'EM | dXmis: {dXmis:1.4f}; rdXmis: {rdXmis:1.4f}')
        for it in pbar:
            pbar.set_description(f'EM | dXmis: {dXmis:1.4f}; rdXmis: {rdXmis:1.4f}')
        # while ((it < maxit) & (rdXmis > self.tol)):
            # it = it + 1
            if rdXmis > self.tol:
                pass
            else:
                print('GraphEM.EM(): Tolerance achieved.')
                break

            CovRes = np.zeros((p,p))
            D = np.sqrt(np.diag(C))
            D[abs(D) < 1e-3] = 1.0  # Do not scale constant variables
            X = X/D
            C = (C/D)/(D[:,None])  # Correlation matrix
            for i1 in range(nb_pattern):
                pm = pattern[i1,:].sum() # Length of the patterm
                if pm > 0:
                    avlr = find(~pattern[i1,:])
                    misr = find(pattern[i1,:])
                    if use_iridge:
                        B, S, __, __ = iridge.iridge(C, avlr, misr, n-1)
                    else: 
                        B, S = ols(C, avlr, misr)
                    ind_obs = find(pattern_ix == i1)
                    mp = len(ind_obs)  # Number of rows matching current pattern
                    Xmis[np.ix_(ind_obs,misr)] = X[np.ix_(ind_obs,avlr)].dot(B)
                    CovRes[np.ix_(misr,misr)] = CovRes[np.ix_(misr,misr)] + mp*S
            # Return to original scaling
            X = X*D
            Xmis = Xmis * D
            C = (C*D)*(D[:,None])
            CovRes = CovRes*D*(D[:,None])
            dXmis = np.linalg.norm(Xmis[indmis] - X[indmis]) / np.sqrt(nmis)
            nXmis_pre  = np.linalg.norm((X+M)[indmis]) / np.sqrt(nmis)
            if nXmis_pre < 1e-16:
                rdXmis   = np.inf
            else:
                rdXmis   = dXmis / nXmis_pre
            X[indmis]  = Xmis[indmis]

            Mup = X.mean(axis=0)
            X = X-Mup
            M = M + Mup
                
            # Re-estimate C
            if use_iridge:
                C = (X.T.dot(X) + CovRes)/(n-1)
            else:
                C = self.fit_Sigma(np.cov(X.T) + CovRes/(n-1), self.graph)
            # if verbose: print("%1.3d     %1.4f     %1.4f" % (it, dXmis, rdXmis))
        X = X + M

        return [X, C, M]
            
    def fit_Sigma(self, S, graph):
        ''' Estimates the covariance matrix of the field using the provided graph
        
        Parameters
        ----------
        S: matrix
            Sample covariance matrix
        graph: matrix
            Adjacency matrix of the graph
            
        Returns
        -------
        C: matrix
            Estimated covariance matrix
        
        '''
        C = fitggm(S, graph)
        return C
        
    def bootstrap(self, N_boot=20, blocksize=2, n_proc=4, save_graphs = False):
        ''' Block bootstrap method for the GraphEM algorithm
        
        Parameters
        ----------
        N_boot: int
            Number of bootstrap samples to use (Default = 20)
        blocksize: int
            Size of the blocks used when constructing bootstrap samples (Default = 2)
        n_proc: int
            Number of processors available for distributed computing (not currently implemented). 
        save_graphs: boolean
            Indicated whether or not to save all the graphs estimated via graph_greedy_search (Default = False)
            
        Returns
        -------
        self.field_r_all: matrix (sample x time x space)
            Matrix containing the reconstructed climate field for each bootstrap sample
        
        '''
        field = np.copy(self.field)
        proxy = np.copy(self.proxy)

        [n,pt] = self.field.shape
        verif = np.setdiff1d(range(n),self.calib)
        
        np.random.seed()
        field_r_all = np.zeros((N_boot,n,pt))
        
        # Build bootstrap indices
        n_calib = len(self.calib)
        n_verif = len(verif)
        
        min_calib = min(self.calib)
        max_calib = max(self.calib)
        min_verif = min(verif)
        max_verif = max(verif)
        
        nb_blocks_calib = np.int(math.ceil(np.double(n_calib)/blocksize))
        nb_blocks_verif = np.int(math.ceil(np.double(n_verif)/blocksize))
        
        last_block_size_calib = np.int(n_calib - (nb_blocks_calib-1)*blocksize)
        last_block_size_verif = np.int(n_verif - (nb_blocks_verif-1)*blocksize)
        
        bootindices = []

        for i1 in range(N_boot):
            # Pick indices for the beginning of the blocks
            block_tmp_calib = np.random.choice(range(min_calib, max_calib-blocksize+1), nb_blocks_calib, replace=True)
            block_tmp_verif = np.random.choice(range(min_verif, max_verif-blocksize+1), nb_blocks_verif, replace=True)
            
            # Construct blocks
            block_tmp_c = [range(i,i+blocksize) for i in block_tmp_calib[:-1]]
            block_tmp_v = [range(i,i+blocksize) for i in block_tmp_verif[:-1]]
            # Add last blocks (that may be of a different size if the calib/verif periods are not divisible by blocksize)
            block_tmp_c.extend([range(block_tmp_calib[-1],block_tmp_calib[-1]+last_block_size_calib)])
            block_tmp_v.extend([range(block_tmp_verif[-1],block_tmp_verif[-1]+last_block_size_verif)])
        
            block_c = [i for subblock in block_tmp_c for i in subblock]        
            block_v = [i for subblock in block_tmp_v for i in subblock]        
            
            block = np.zeros(n, dtype=np.int)
            block[self.calib] = block_c
            block[verif] = block_v

            bootindices.append(block)
        
        self.bootindices = bootindices
        G = np.eye(self.field.shape[1]+self.proxy.shape[1])

        self.boot_graphs = []
            
        # Begin bootstrap
        for i1 in range(N_boot):
            self.fit(field[bootindices[i1],:], proxy[bootindices[i1],:], self.calib, [], self.lonlat, self.sp_FF, self.sp_FP, self.sp_PP, N_graph = self.N_graph, bootstrap=False, cutoff_radius = self.cutoff_radius, graph_method = self.graph_method)
            if save_graphs:
                self.boot_graphs.append(self.graph)
            C0 = self.Sigma
            M0 = self.mu
            # Imputation step (no graph estimation)
            self.fit(field, proxy, self.calib, G, self.lonlat, self.sp_FF, self.sp_FP, self.sp_PP, C0 = C0, M0 = M0, maxit=1, graph_method = self.graph_method, cutoff_radius = self.cutoff_radius, estimate_graph = False)
            field_r_all[i1,:,:] = self.field_r
        self.field_r_all = field_r_all
            

def adj_matrix(S, tol=1e-3):
    '''
    Estimates an adjacency matrix via thresholding
    
    Entries larger than 'tol' are set to 1, and to 0 otherwise. 
    Diagonal entries are set to 0. 
    
    Parameters
    ----------
    S: matrix
        original matrix
    tol: float
        threshold (Default = 1e-3)
        
    Returns
    -------
    adj: matrix
        Adjacency matrix
    '''
    adj = (abs(S) > tol).astype(int)
    adj = adj - np.diag(np.diag(adj))
    return adj

def sp_level_3(O, ind_F, ind_P):
    '''
    Computes the sparsity level (percentage of nonzero entries) 
    of a matrix in the in-field, 
    field/proxy, and proxy/proxy part of the matrix O.
    
    Parameters
    ----------
    O: matrix
        Matrix whose sparsity level is sought (real, symmetric)
    ind_F: vector of int
        Indices of the field locations
    ind_P: vector of int
        Indices of the proxy locations
        
    Returns
    -------
    level_TT: sparsity level of the in-field block of the matrix
    level_TP: sparsity level of the field/proxy block of the matrix
    level_PP: sparsity level of the proxy/proxy block of the matrix
    adj: adjacency matrix associated to the matrix O
    '''
    p = O.shape[0]
    d = np.diag(O)
    dinv = np.sqrt(d)**(-1)
    D = np.diag(dinv)
    O_scaled = D.dot(O).dot(D)
    adj = adj_matrix(O_scaled)
    p_TT = len(ind_F)
    p_PP = len(ind_P)

    adj_TT = adj[np.ix_(ind_F, ind_F)]
    level_TT = adj_TT.sum().astype(float) / (p_TT*(p_TT-1))*100

    adj_TP = adj[np.ix_(ind_F, ind_P)]
    level_TP = adj_TP.sum().astype(float) / (p_TT*p_PP)*100

    adj_PP = adj[np.ix_(ind_P, ind_P)]
    level_PP = adj_PP.sum().astype(float) / (p_PP*(p_PP-1))*100

    adj = adj + np.eye(p)
    return [level_TT, level_TP, level_PP, adj]


def fitggm(S, graph, tol=5e-3, maxit=200):
    '''
    Estimates the covariance matrix of the Gaussian graphical model via 
    maximum likelihood
    
    Parameters
    ----------
    S: matrix (p x p)
        Sample covariance matrix of the field
    graph: matrix (p x p)
        Adjacency matrix of the graph
        
    tol: float
        Convergence tolerence (Default = 5e-3)
    maxit: int
        Maximum number of iterations of the algorithm (Default = 200)
        
    Returns
    -------
    W: matrix (p x p)
        Estimated covariance matrix
    '''
    p = S.shape[0]
    pp = p*(p-1)/2
    beta = np.zeros((p,p))

    W = np.copy(S)
    adj = np.copy(graph)

    for it in range(maxit):
        W_old = np.copy(W)
        for i1 in range(p):
            ind = np.copy(adj[i1,:])
            ind[i1] = 0
            indices = find(ind)
            W11_star = W[np.ix_(indices,indices)]
            
            # DEBUG ONLY:
            #c = np.linalg.cond(W11_star) 
            #plt.imshow(W11_star)
            #plt.title(r'$\Omega_{11}$ for i = ' +str(i1) +', condition number = {:6.3f}'.format(c))
            
            s12_star = S[indices,i1]
            if ~np.all(ind==0):
                #beta_star = np.linalg.solve(W11_star,s12_star)
                beta_star = linalg.solve(W11_star,s12_star, assume_a='sym')
                beta[i1,:] = 0
                beta[i1,indices] = beta_star
            else:
                beta[i1,:] = 0
            W[:,i1] = W.dot(beta[i1,:])
            W[i1,:] = W[:,i1]
            W[i1,i1] = S[i1,i1]
        Var = W - W_old
        np.fill_diagonal(Var,0)
        Var = np.triu(Var)
        meanvariation = abs(Var).sum()/pp
        W_old2 = np.copy(W_old)
        np.fill_diagonal(W_old2,0)
        W_old2 = np.triu(W_old2)
        mean_old = abs(W_old2).sum()/pp
        if meanvariation <= tol * mean_old:
            # Compute inverse covariance if desired
            #for i1 in range(p):
            #    O22 = 1/(S[i1,i1]-W[i1,:].dot(beta[i1,:]))
            #    Omega[i1,:] = -1*beta[i1,:]*O22
            #    Omega[:,i1] = Omega[i1,:]
            #    Omega[i1,i1] = O22
            return W
    print("Warning: fitggm has reached its maximum number of iterations.")
    return W


def ols(Sigma, xind, yind):
    '''
    Computes regression coefficients in a multivariate Gaussian model
    
    Parameters
    ----------
    Sigma: matrix (p x p) 
        Covariance matrix of the field
    xind: vector of int 
        Indices of x
    yind: vector of int
        Indices of y
        
    Returns
    -------
    B: matrix
        Regression coefficients
    S: matrix
        Covariance of the residuals
    '''
    B = linalg.inv(Sigma[np.ix_(xind,xind)]).dot(Sigma[np.ix_(xind,yind)])
    S = Sigma[np.ix_(yind,yind)] - Sigma[np.ix_(yind, xind)].dot(B)
    return [B, S]


def unique_rows(a):
    '''
    Returns the unique rows of a matrix
    
    Parameters
    ----------
    a: matrix
    
    Returns
    -------
    unique_a: matrix
        Matrix containing the unique rows of a
    idx: Vector indicating the unique rows of a
    '''
    b = np.ascontiguousarray(a)
    (uniq,idx) = np.unique(b.view(np.dtype((np.void, b.dtype.itemsize*b.shape[1]))),return_inverse = True)
    unique_a = uniq.view(b.dtype).reshape(-1, b.shape[1])
    return (unique_a,idx)
    
class verif_stats:
    '''
    Class to compute verification statistics of a reconstructed climate field
    
    Attributes
    ----------
    T_hat: matrix (n x p (no bootstrap) or N x n x p (bootstrap)) 
        Reconstructed variable
    T: matrix(n x p)
        Target variable
    calib: vector of int
        Indices of the calibration period (verification inferred as the complement)
        
    Usage
    -----
    Statistics are computed at creation of the object and are stored in 
    the MSE, RE, CE, and R2 attributes. 
    
    If T_hat is an N x n x p matrix 
    (i.e., N reconstructions are provided), the attribute 
    "XXX_all" is a matrix (the i-th row is the value of the statistic for the 
    i-th reconstruction), "XXX" contains the value of the statistic averaged over 
    all reconstructions, and "XXX_std" contains the standard deviation of the 
    statistics ("XXX" = MSE, RE, CE, or R2).  
    
    Methods
    -------
    _MSE: mean squared error
    _RE: Reduction of error
    _CE: Coefficient of efficiency
    _R2: Coefficient of determination (R^2)
    stats: computes all the statistics
    __str__: displays the MSE, RE, CE, and R2
    '''
    def __init__(self, field_recon, field_target, calib):
        '''
        Initializes the verif_stats object
        
        Parameters
        ----------
        field_recon: matrix (n x p (no bootstrap) or N x n x p (bootstrap))
            Reconstructed climate field
        field_target: matrix (n x p)
            Target climate field
        calib: vector of int
            Indices of the calibration period
        '''
        self.T_hat = field_recon
        self.T     = field_target
        self.calib = calib
        self.stats()

    def _MSE(self, T, T_hat, calib):
        '''
        Computes the mean squared error (MSE) of the reconstruction 
        '''
        [n,p] = T.shape
        verif = np.setdiff1d(range(n),calib)
        Tv = T[verif,:]
        Tv_hat = T_hat[verif,:]
        return ((Tv - Tv_hat)**2).mean(axis=0)
        
    def _RE(self, T, T_hat, calib):
        '''
        Computes the reduction of error (RE) of the reconstruction
        '''
        [n,p] = T.shape
        verif = np.setdiff1d(range(n),calib)
        Tv = T[verif,:]
        Tv_hat = T_hat[verif,:]
        
        Tbar_calib = T[calib,:].mean(axis=0)
        
        RE = 1 - ((Tv-Tv_hat)**2).sum(axis=0) / ((Tv-Tbar_calib)**2).sum(axis=0)
        return RE
        
    def _CE(self, T, T_hat, calib):
        '''
        Computes the coefficient of efficiency (CE) of the reconstruction
        '''
        [n,p] = T.shape
        verif = np.setdiff1d(range(n),calib)
        Tv = T[verif,:]
        Tv_hat = T_hat[verif,:]
        
        Tbar_verif = T[verif,:].mean(axis=0)
        
        CE = 1 - ((Tv-Tv_hat)**2).sum(axis=0) / ((Tv-Tbar_verif)**2).sum(axis=0)
        return CE

    def _R2(self, T, T_hat, calib):
        '''
        Computes the coefficient of determination (R^2) of the reconstruction
        '''
        [n,p] = T.shape
        verif = np.setdiff1d(range(n),calib)
        Tv = T[verif,:]
        Tv_hat = T_hat[verif,:]
        R2 = np.zeros(p)
        
        for i1 in range(p):
            rho = np.corrcoef(Tv[:,i1],Tv_hat[:,i1])  
            R2[i1]=rho[0,1]**2
            
        return R2

    def stats(self):
        '''
        Computes the MSE, RE, CE, and R^2 statistics. 
        '''
        if len(self.T_hat.shape) == 3:
            # bootstrap 
            self.MSE_all = np.zeros((self.T_hat.shape[0],self.T_hat.shape[2]))
            self.RE_all = np.zeros((self.T_hat.shape[0],self.T_hat.shape[2]))
            self.CE_all = np.zeros((self.T_hat.shape[0],self.T_hat.shape[2]))
            self.R2_all = np.zeros((self.T_hat.shape[0],self.T_hat.shape[2]))
            
            for i1 in range(self.T_hat.shape[0]):
                self.MSE_all[i1,:] = self._MSE(self.T,self.T_hat[i1,:,:],self.calib)
                self.RE_all[i1,:] = self._RE(self.T,self.T_hat[i1,:,:],self.calib)
                self.CE_all[i1,:] = self._CE(self.T,self.T_hat[i1,:,:],self.calib)
                self.R2_all[i1,:] = self._R2(self.T,self.T_hat[i1,:,:],self.calib)
        
            self.MSE = self.MSE_all.mean(axis=0)
            self.RE = self.RE_all.mean(axis=0)
            self.CE = self.CE_all.mean(axis=0)
            self.R2 = self.R2_all.mean(axis=0)
            
            self.MSE_std = self.MSE_all.std(axis=0)
            self.RE_std = self.RE_all.std(axis=0)
            self.CE_std = self.CE_all.std(axis=0)
            self.R2_std = self.R2_all.std(axis=0)
            
        else:
            self.MSE = self._MSE(self.T,self.T_hat,self.calib)
            self.RE = self._RE(self.T,self.T_hat,self.calib)
            self.CE = self._CE(self.T,self.T_hat,self.calib)
            self.R2 = self._R2(self.T,self.T_hat,self.calib)
        
    def __str__(self):
        '''
        Displays the calculated values of MSE, RE, CE, and R2
        '''
        return "Mean MSE = %1.4f, Mean RE = %1.4f, Mean CE = %1.4f, Mean R2 = %1.4f" % (self.MSE.mean(), self.RE.mean(), self.CE.mean(), self.R2.mean())    
        
def find(condition):
    '''
    Code to replace the old matplotlib.mlab.find function
    (Return the indices where some condition is true)
    '''
    res, = np.nonzero(np.ravel(condition))
    return res  


class KCV:
    def __init__(self, stats, cutoff_radii, adjs=None):
        self.stats = stats
        self.cutoff_radii = cutoff_radii
        self.adjs = adjs

    def plot(self, figsize=[6, 4], ylabel='MSE', ax=None, verbose=True):
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        ax.plot(self.cutoff_radii, self.stats.mean(axis=0), '-o')
        ax.fill_between(self.cutoff_radii, self.stats.min(axis=0), self.stats.max(axis=0), alpha=0.3)
        ax.set_xlabel('Cutoff radius')
        ax.set_ylabel(ylabel)
        ax.set_title('K-fold Cross Validation')

        if verbose:
            opt_rad = self.cutoff_radii[np.argmin(self.stats.mean(axis=0))]
            print(f'Minimum reached for R = {opt_rad:d} km.')

        if 'fig' in locals():
            return fig, ax
        else:
            return ax

    def plot_adjs(self, figsize=(10, 20)):
        fig = plt.figure(figsize=figsize)
        nr = len(self.cutoff_radii)
        nf = len(self.adjs)//nr
        gs = gridspec.GridSpec(nf, nr)
        gs.update(wspace=0.5, hspace=0.5)
        ax = {}
        i = 0
        for k, v in self.adjs.items():
            fold, param = k
            ax[k] = fig.add_subplot(gs[i])
            ax[k] = v.plot_adj(title=f'Fold: {fold}, Parameter: {param}', ax=ax[k])
            i += 1

        return fig, ax
        

    def one_sd_rule(self, objective_fun=None, param_vector=None, verbose=True):
        ''' One SD rule: selects the least complex model whose objective function (e.g. MSE) is 
        within 1 standard deviation (fold-wise) of the global minimum.
    
        Parameters
        ----------
        objective_fun : array-like (NumPy, Pandas or similar)
            values of the objective function, with the first dimension (0) assumed as the sample dimension
        param_vector :  array-like (NumPy, Pandas or similar)
            values of the control parameter used to compute the objective function. 

        Returns
        -------
        opt_param, the optimal parameter value
        '''    
        objective_fun = self.stats if objective_fun is None else objective_fun
        param_vector = self.cutoff_radii if param_vector is None else param_vector

        assert objective_fun.shape[1] == param_vector.shape[0], "param_vector has an incongruent dimension. See docstring"
    
        obj_mean = objective_fun.mean(axis=0)
        obj_min = obj_mean.min()
        i_min = np.argmin(obj_mean)
        sd = np.std(objective_fun,axis=0)[i_min]
        idx = np.where(obj_mean <= obj_min + sd)[0][0] # take first value
        opt_param = param_vector[idx]
        if verbose:
            print(f'The 1SD rule selects R = {opt_param:d} km.')
    
        return opt_param 


import os
from setuptools import setup, find_packages
from distutils.core import setup, Extension
from distutils import sysconfig
from Cython.Distutils import build_ext

with open('README.rst', 'r') as fh:
    long_description = fh.read()

quiclib = Extension(
    'quiclib', sources=['./graphem/QUIC.cpp'],
    extra_link_args=['-llapack', '-lblas', '-lstdc++', '-fPIC'],
)

class NoSuffixBuilder(build_ext):
    def get_ext_filename(self, ext_name):
        filename = super().get_ext_filename(ext_name)
        suffix = sysconfig.get_config_var('EXT_SUFFIX')
        ext = os.path.splitext(filename)[1]
        return os.path.join('./graphem', filename.replace(suffix, "")+ext)

setup(
    name='cfr-graphem',
    version='0.2.9',
    description='The GraphEM component for CFR',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author='Feng Zhu, Julien Emile-Geay',
    author_email='fengzhu@ucar.edu, julieneg@usc.edu',
    url='https://github.com/fzhu2e/cfr-graphem',
    packages=find_packages(),
    include_package_data=True,
    license='BSD 3-Clause',
    zip_safe=False,
    cmdclass={'build_ext': NoSuffixBuilder},
    ext_modules=[quiclib],
    keywords='climate field reconstruction',
    classifiers=[
        'Natural Language :: English',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    install_requires=[
        'cfr',
    ],
)



# compile C++ file with (Linux-compatible): g++ -shared QUIC.cpp -o quiclib.so -fPIC -llapack -lblas -lstdc++

import numpy as np
import os
from ctypes import CDLL, c_double, c_int


lib_path = os.path.dirname(os.path.abspath(__file__))
quiclib = CDLL(os.path.join(lib_path,'quiclib.so'))

def QUIC(Si, rho):
	'''
	Solves the graphical lasso problem using QUIC.
	(Wrapper for the C++ code QUIC.cpp.)
	Reference: Hsieh et al., "QUIC: Quadratic Approximation for Sparse Inverse
		Covariance Estimation", Journal of Machine Learning Research 15 (2014) 2911-2947. 
	Note: In order to use the code, the C++ code QUIC.cpp needs to be compiled first. 

	Parameters:
	-----------
	Si: matrix
		Sample covariance matrix
	rho: float
		Regularization parameter
	Returns:
	--------
	Sigma: matrix
		Estimated covariance matrix
	Omega: matrix
		Estimated inverse covariance matrix
	'''
	d1 = Si.shape[0]
	d = d1**2
	Si = Si.reshape(d)
	rho = rho.reshape(d)
	Sigma = np.zeros(d)
	Omega = np.zeros(d)
	S = (c_double*d)()
	rhomat = (c_double*d)()
	O = (c_double*d)()
	W = (c_double*d)()
	Id = np.eye(d1).reshape(d)
	for i1 in range(d):
		S[i1] = (c_double)(Si[i1])
		rhomat[i1] = (c_double)(rho[i1])
		W[i1] = (c_double)(Id[i1])
		O[i1] = (c_double)(Id[i1])
		
	n = (c_int*1)(d1)
	npath = (c_int*1)(1)
	path = (c_int*1)(0)
	tol = (c_double*1)(1e-3)
	msg = (c_int*1)(0)
	maxIter = (c_int*1)(1000)
	opt = (c_double*1)(0)
	cputime = (c_double*1)(0)
	it = (c_int*1)(0)
	dGap = (c_double*1)(0)

	quiclib.QUIC('d',n,S,rhomat,npath,path,tol,msg,maxIter,O,W,opt,cputime,it,dGap)

	for i1 in range(d):
		Sigma[i1] = W[i1]
		Omega[i1] = O[i1]
	Sigma = Sigma.reshape((d1,d1))
	Omega = Omega.reshape((d1,d1))
	return [Sigma, Omega]
    
